@@ -60,6 +60,7 @@
 #include "mod_proxy.h"
 #include "http_main.h"
 #include "http_log.h"
+#include "http_core.h"
 
 #define AUTODETECT_PWD
 
@@ -101,11 +102,7 @@ static int ftp_check_string(const char *x)
 	    ch = ap_proxy_hex2c(&x[i + 1]);
 	    i += 2;
 	}
-#ifndef CHARSET_EBCDIC
-	if (ch == '\015' || ch == '\012' || (ch & 0x80))
-#else /*CHARSET_EBCDIC*/
-	if (ch == '\r' || ch == '\n' || (os_toascii[ch] & 0x80))
-#endif /*CHARSET_EBCDIC*/
+	if (ch == CR || ch == LF || (OS_ASC(ch) & 0x80))
 	    return 0;
     }
     return 1;
@@ -139,7 +136,8 @@ int ap_proxy_ftp_canon(request_rec *r, char *url)
     strp = strchr(url, ';');
     if (strp != NULL) {
 	*(strp++) = '\0';
-	parms = ap_proxy_canonenc(p, strp, strlen(strp), enc_parm, r->proxyreq);
+	parms = ap_proxy_canonenc(p, strp, strlen(strp), enc_parm,
+				  r->proxyreq);
 	if (parms == NULL)
 	    return HTTP_BAD_REQUEST;
     }
@@ -152,15 +150,15 @@ int ap_proxy_ftp_canon(request_rec *r, char *url)
     if (!ftp_check_string(path))
 	return HTTP_BAD_REQUEST;
 
-    if (!r->proxyreq && r->args != NULL) {
+    if (r->proxyreq == NOT_PROXY && r->args != NULL) {
 	if (strp != NULL) {
-	    strp = ap_proxy_canonenc(p, r->args, strlen(r->args), enc_parm, 1);
+	    strp = ap_proxy_canonenc(p, r->args, strlen(r->args), enc_parm, STD_PROXY);
 	    if (strp == NULL)
 		return HTTP_BAD_REQUEST;
 	    parms = ap_pstrcat(p, parms, "?", strp, NULL);
 	}
 	else {
-	    strp = ap_proxy_canonenc(p, r->args, strlen(r->args), enc_fpath, 1);
+	    strp = ap_proxy_canonenc(p, r->args, strlen(r->args), enc_fpath, STD_PROXY);
 	    if (strp == NULL)
 		return HTTP_BAD_REQUEST;
 	    path = ap_pstrcat(p, path, "?", strp, NULL);
@@ -290,7 +288,7 @@ static long int send_dir(BUFF *f, request_rec *r, cache_req *c, char *cwd)
 	path[n-1] = '\0';
 
     /* print "ftp://host/" */
-    n = ap_snprintf(buf, sizeof(buf), "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">\n"
+    n = ap_snprintf(buf, sizeof(buf), DOCTYPE_HTML_3_2
 		"<HTML><HEAD><TITLE>%s%s</TITLE>\n"
 		"<BASE HREF=\"%s%s\"></HEAD>\n"
 		"<BODY><H2>Directory of "
@@ -322,8 +320,11 @@ static long int send_dir(BUFF *f, request_rec *r, cache_req *c, char *cwd)
     while (!con->aborted) {
 	n = ap_bgets(buf, sizeof buf, f);
 	if (n == -1) {		/* input error */
-	    if (c != NULL)
+	    if (c != NULL) {
+		ap_log_rerror(APLOG_MARK, APLOG_ERR, c->req,
+		    "proxy: error reading from %s", c->url);
 		c = ap_proxy_cache_error(c);
+	    }
 	    break;
 	}
 	if (n == 0)
@@ -382,8 +383,11 @@ static long int send_dir(BUFF *f, request_rec *r, cache_req *c, char *cwd)
 	o = 0;
 	total_bytes_sent += n;
 
-	if (c != NULL && c->fp && ap_bwrite(c->fp, buf, n) != n)
+	if (c != NULL && c->fp && ap_bwrite(c->fp, buf, n) != n) {
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, c->req,
+		"proxy: error writing to %s", c->tempfile);
 	    c = ap_proxy_cache_error(c);
+	}
 
 	while (n && !r->connection->aborted) {
 	    w = ap_bwrite(con->client, &buf[o], n);
@@ -413,7 +417,7 @@ static long int send_dir(BUFF *f, request_rec *r, cache_req *c, char *cwd)
  */
 static int ftp_unauthorized (request_rec *r, int log_it)
 {
-    r->proxyreq = 0;
+    r->proxyreq = NOT_PROXY;
     /* Log failed requests if they supplied a password
      * (log username/password guessing attempts)
      */
@@ -503,7 +507,7 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
      */
     if ((password = ap_table_get(r->headers_in, "Authorization")) != NULL
 	&& strcasecmp(ap_getword(r->pool, &password, ' '), "Basic") == 0
-	&& (password = ap_uudecode(r->pool, password))[0] != ':') {
+	&& (password = ap_pbase64decode(r->pool, password))[0] != ':') {
 	/* Note that this allocation has to be made from r->connection->pool
 	 * because it has the lifetime of the connection.  The other allocations
 	 * are temporary and can be tossed away any time.
@@ -557,6 +561,7 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
 	return HTTP_INTERNAL_SERVER_ERROR;
     }
 
+#if !defined(TPF) && !defined(BEOS)
     if (conf->recv_buffer_size > 0
 	&& setsockopt(sock, SOL_SOCKET, SO_RCVBUF,
 		       (const char *) &conf->recv_buffer_size, sizeof(int))
@@ -564,6 +569,7 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
 	    ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
 			 "setsockopt(SO_RCVBUF): Failed to set ProxyReceiveBufferSize, using default");
     }
+#endif
 
     if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *) &one,
 		   sizeof(one)) == -1) {
@@ -771,7 +777,7 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
 
     if (parms[0] != 'a') {
 	/* set type to image */
-	/* TM - Added \015\012 to the end of TYPE I, otherwise it hangs the
+	/* TM - Added CRLF to the end of TYPE I, otherwise it hangs the
 	   connection */
 	ap_bputs("TYPE I" CRLF, f);
 	ap_bflush(f);
@@ -809,6 +815,7 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
 	return HTTP_INTERNAL_SERVER_ERROR;
     }
 
+#if !defined (TPF) && !defined(BEOS)
     if (conf->recv_buffer_size) {
 	if (setsockopt(dsock, SOL_SOCKET, SO_RCVBUF,
 	       (const char *) &conf->recv_buffer_size, sizeof(int)) == -1) {
@@ -816,6 +823,7 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
 			 "setsockopt(SO_RCVBUF): Failed to set ProxyReceiveBufferSize, using default");
 	}
     }
+#endif
 
     ap_bputs("PASV" CRLF, f);
     ap_bflush(f);
@@ -1124,15 +1132,18 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
     resp_hdrs = ap_make_table(p, 2);
     c->hdrs = resp_hdrs;
 
+    ap_table_setn(resp_hdrs, "Date", ap_gm_timestr_822(r->pool, r->request_time));
+    ap_table_setn(resp_hdrs, "Server", ap_get_server_version());
+
     if (parms[0] == 'd')
-	ap_table_set(resp_hdrs, "Content-Type", "text/html");
+	ap_table_setn(resp_hdrs, "Content-Type", "text/html");
     else {
 	if (r->content_type != NULL) {
-	    ap_table_set(resp_hdrs, "Content-Type", r->content_type);
+	    ap_table_setn(resp_hdrs, "Content-Type", r->content_type);
 	    Explain1("FTP: Content-Type set to %s", r->content_type);
 	}
 	else {
-	    ap_table_set(resp_hdrs, "Content-Type", "text/plain");
+	    ap_table_setn(resp_hdrs, "Content-Type", ap_default_type(r));
 	}
 	if (parms[0] != 'a' && size != NULL) {
 	    /* We "trust" the ftp server to really serve (size) bytes... */
@@ -1140,6 +1151,10 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
 	    Explain1("FTP: Content-Length set to %s", size);
 	}
     }
+    if (r->content_encoding != NULL && r->content_encoding[0] != '\0') {
+	Explain1("FTP: Content-Encoding set to %s", r->content_encoding);
+	ap_table_setn(resp_hdrs, "Content-Encoding", r->content_encoding);
+    }
 
 /* check if NoCache directive on this host */
     for (i = 0; i < conf->nocaches->nelts; i++) {
@@ -1182,18 +1197,17 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
 	ap_bpushfd(data, dsock, dsock);
     }
 
-#ifdef CHARSET_EBCDIC
-/*    bsetflag(data, B_ASCII2EBCDIC|B_EBCDIC2ASCII, 0);*/
-#endif /*CHARSET_EBCDIC*/
-
     ap_hard_timeout("proxy receive", r);
 /* send response */
 /* write status line */
     if (!r->assbackwards)
 	ap_rvputs(r, "HTTP/1.0 ", r->status_line, CRLF, NULL);
     if (c != NULL && c->fp != NULL
-	&& ap_bvputs(c->fp, "HTTP/1.0 ", r->status_line, CRLF, NULL) == -1)
-	c = ap_proxy_cache_error(c);
+	&& ap_bvputs(c->fp, "HTTP/1.0 ", r->status_line, CRLF, NULL) == -1) {
+	    ap_log_rerror(APLOG_MARK, APLOG_ERR, c->req,
+		"proxy: error writing CRLF to %s", c->tempfile);
+	    c = ap_proxy_cache_error(c);
+    }
 
 /* send headers */
     tdo.req = r;
@@ -1202,8 +1216,11 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
 
     if (!r->assbackwards)
 	ap_rputs(CRLF, r);
-    if (c != NULL && c->fp != NULL && ap_bputs(CRLF, c->fp) == -1)
+    if (c != NULL && c->fp != NULL && ap_bputs(CRLF, c->fp) == -1) {
+	ap_log_rerror(APLOG_MARK, APLOG_ERR, c->req,
+	    "proxy: error writing CRLF to %s", c->tempfile);
 	c = ap_proxy_cache_error(c);
+    }
 
     ap_bsetopt(r->connection->client, BO_BYTECT, &zero);
     r->sent_bodyct = 1;
@@ -1222,6 +1239,7 @@ int ap_proxy_ftp_handler(request_rec *r, cache_req *c, char *url)
 
 	/* XXX: we checked for 125||150||226||250 above. This is redundant. */
 	if (rc != 226 && rc != 250)
+            /* XXX: we no longer log an "error writing to c->tempfile" - should we? */
 	    c = ap_proxy_cache_error(c);
     }
     else {
